<?php

namespace Drupal\http_client_logger;

use Drupal\Core\Config\ConfigFactory;
use Drupal\http_client_logger\Form\AdminConfigurationForm;
use Drupal\http_client_logger\GuzzleHttp\MessageFormatter;
use GuzzleHttp\Exception\GuzzleException;
use GuzzleHttp\Middleware;
use GuzzleHttp\Promise\PromiseInterface;
use Monolog\Formatter\LineFormatter;
use Monolog\Handler\StreamHandler;
use Monolog\Logger;
use Psr\Http\Message\RequestInterface;
use Psr\Http\Message\ResponseInterface;
use Psr\Log\LogLevel;

/**
 * Attach a middleware which can log HTTP requests and responses.
 *
 * "Modules wishing to add additional middlewares to the handler stack can
 *  create services and tag them as http_client_middleware. Each service must
 *  contain an __invoke method that returns a closure which will serve as the
 *  middleware."
 */
class HttpClientMiddlewareLogger {

  /**
   * The name of the logger-channel.
   *
   * @var string
   */
  const LOGGER_CHANNEL_NAME = 'HttpClientLogger';

  /**
   * Default monolog LineFormatter definition.
   *
   * @var array
   */
  const MONOLOG_LINEFORMATTER_DEFAULT = ["%message%\n", NULL, TRUE, TRUE];
  /**
   * Module configuration.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected $config;

  /**
   * Constructor.
   *
   * @param \Drupal\Core\Config\ConfigFactory $config_factory
   *   The Config Factory service.
   */
  public function __construct(ConfigFactory $config_factory) {
    $this->config = $config_factory->get(AdminConfigurationForm::CONFIG_KEY);
  }

  /**
   * Middleware to log requests and responses.
   *
   * @see \Drupal\Core\Http\HandlerStackConfigurator
   * @see \GuzzleHttp\Middleware:log()
   */
  public function __invoke() {
    $tap = ($this->config->get('enabled')) ? [$this, 'log'] : NULL;
    // Tap the 'after' method, when the response is available.
    return Middleware::tap(NULL, $tap);
  }

  /**
   * Log a request and/or response.
   *
   * @param \Psr\Http\Message\RequestInterface $request
   *   The HTTP request.
   * @param array $options
   *   Options passed to the Guzzle HTTP client.
   * @param \GuzzleHttp\Promise\PromiseInterface $promise
   *   A promise which fulfils with a \GuzzleHttp\Psr7\Response.
   *
   * @see \GuzzleHttp\Middleware::tap()
   */
  public function log(RequestInterface $request, array $options, PromiseInterface $promise) {
    $this->logRequest($request);
    $promise->then(
      // Promise fulfilled.
      [$this, 'logResponse'],
      // Promise rejected.
      [$this, 'logError']);
  }

  /**
   * Log a HTTP request.
   *
   * @param \GuzzleHttp\Psr7\RequestInterface $request
   *   The HTTP request.
   */
  public function logRequest(RequestInterface $request) {
    // Attach the request so that it's available to the response logger.
    $this->request = $request;
    if ($filename = $this->destinationFilename('request')) {
      // @todo Implement tokenisation for log destination filenames.
      // @todo Make the recording message format configurable.
      $logger = $this->fileLogger($filename);
      $guzzleFormatter = new MessageFormatter(MessageFormatter::REQUEST_DEFAULT);
      $message = $guzzleFormatter->format($request);
      $logger->log(LogLevel::INFO, $message);
    }
  }

  /**
   * Log a HTTP response.
   *
   * @param \GuzzleHttp\Psr7\ResponseInterface $response
   *   The HTTP response.
   */
  public function logResponse(ResponseInterface $response) {
    if ($filename = $this->destinationFilename('response')) {
      $logger = $this->fileLogger($filename);
      $guzzleFormatter = new MessageFormatter(MessageFormatter::RESPONSE_DEFAULT);
      $message = $guzzleFormatter->format($this->request, $response);
      $logger->log(LogLevel::INFO, $message);
    }
  }

  /**
   * Log an error.
   *
   * @param \GuzzleHttp\Exception\GuzzleException $error
   *   Error raised by Guzzle.
   */
  public function logError(GuzzleException $error) {
    if ($filename = $this->destinationFilename('error')) {
      $logger = $this->fileLogger($filename);
      $guzzleFormatter = new MessageFormatter(MessageFormatter::ERROR_DEFAULT);
      $message = $guzzleFormatter->format($this->request, NULL, $error);
      $logger->log(LogLevel::WARNING, $message);
    }
  }

  /**
   * Create a file-logger to log a message and no additional data.
   *
   * @param mixed $stream
   *   Either a filename, or a stream wrapper.
   *
   * @return \Psr\Log\LoggerInterface
   *   A file-system logger conforming with the PSR specification.
   */
  protected function fileLogger($stream) {
    $loggerHandler = new StreamHandler($stream, Logger::DEBUG);
    // This logger format logs the message and nothing else.
    $loggerHandler->setFormatter(new LineFormatter(...self::MONOLOG_LINEFORMATTER_DEFAULT));
    $logger = new Logger(self::LOGGER_CHANNEL_NAME);
    $logger->pushHandler($loggerHandler);
    return $logger;
  }

  /**
   * Get the filename of the destination log file.
   *
   * @param string $log_type
   *   One of "request", "response", or "error".
   *
   * @return string
   *   The destination filename for the log-type, before any token-replacement
   *   has happened.
   */
  protected function destinationFilename($log_type) {
    if ($this->config->get('destination') === 'common') {
      return $this->config->get('filenames')['common'];
    }
    else {
      return $this->config->get('filenames')[$log_type];
    }
  }

}
